/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** project:
**    OpenGL/cairo anti-aliasing trick
**
** file:
**    cairo-utils.h
**
** author:
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**
** copyright (C) Mirco Mueller, December 2006, placed under the terms of the GPL
**
*******************************************************************************/

#ifndef _CAIRO_UTILS_H
#define _CAIRO_UTILS_H

#include <gtk/gtk.h>
#include <cairo.h>

#include "opengl.h"

void
cairo_draw_round_rect (cairo_t* pContext,
		       double fAspectRatio,  /* aspect-ratio            */
		       double fX,            /* top-left corner         */
		       double fY,            /* top-left corner         */
		       double fCornerRadius, /* "size" of the corners   */
		       double fWidth,        /* width of the rectangle  */
		       double fHeight        /* height of the rectangle */);

void
compute_1d_kernel (gint	   iRadius,    /* radius of kernel                  */
		   gfloat  fDeviation, /* deviation to use, 0.0f for auto   */
		   gfloat* pfKernel    /* pointer to array for coefficients */);

gdouble*
kernel_1d_new (gint    iRadius,	  /* radius of kernel                                */
	       gdouble fDeviation /* deviation to use, pass 0.0f for auto-generation */);

void
kernel_1d_delete (gdouble* pfKernel);

void
convolute (guchar*  pucImageBuffer, /* pointer to image-data */
	   gint	    iImageWidth,    /* width of image        */
	   gint	    iImageHeight,   /* height of image       */
	   gint	    iImageStride,   /* stride of image       */
	   gdouble* pfKernel        /* the kernel            */);

void
glow (guchar* pucSrcBuffer, /* pointer to source             */
      gint    iSrcWidth,    /* width of source               */
      gint    iSrcHeight,   /* height of source              */
      gint    iSrcChannels, /* color-channels in source      */
      guchar* pucDstBuffer, /* pointer to destination        */
      gint*   piDstWidth,   /* pointer to destination width  */
      gint*   piDstHeight   /* pointer to destination height */);

void
blur (guchar* pucSrc,    /* pointer to image-data   */
      gint    iWidth,    /* width of image          */
      gint    iHeight,   /* height of image         */
      gint    iChannels, /* color-channels in image */
      gint    iRadius    /* the blur-radius         */);

gdouble*
build_image_sat (guchar* pucSrcBuffer /* pointer to source image        */,
		 gint	 iWidth	      /* width of source image          */,
		 gint	 iHeight      /* height of source image         */,
		 gint	 iChannels    /* color-channels in source image */);

void
demultiply_alpha (guchar* pucPixelBuffer, /* pointer to image-data */
		  gint iWidth,            /* width of image        */
		  gint iHeight,           /* height of image       */
		  gint iStride            /* stride of image       */);

void
cairo_trim_image (guchar* pucPixelBuffer, /* pointer to image-data */
		  gint iWidth,            /* width of image        */
		  gint iHeight,           /* height of image       */
		  gint iStride            /* stride of image       */);

Texture*
cairo_text_to_texture (gchar* pcText /* the text to render */);

Texture*
cairo_logos_to_texture (gchar* pcSvgFileOne, /* full path first SVG-file  */
			gchar* pcSvgFileTwo, /* full path second SVG-file */
			gint iWidth,	     /* width of window           */
			gint iHeight	     /* height of window          */);

#endif /* _CAIRO_UTILS_H */

