/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** project:
**    OpenGL/cairo anti-aliasing trick
**
** file:
**    events.c
**
** author:
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**
** copyright (C) Mirco Mueller, December 2006, placed under the terms of the GPL
**
*******************************************************************************/

#include <gtk/gtk.h>
#include <gtk/gtkgl.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
#include <GL/gl.h>

#include "opengl.h"

gboolean g_bDraw = TRUE;
gfloat	 g_fFactor = 1.0f;

gboolean
delete_handler (GtkWidget* pWidget,
		GdkEvent* pEvent,
		gpointer data)
{
	opengl_cleanup ();
	gtk_main_quit ();

	return TRUE;
}

void
realize_handler (GtkWidget* pWidget,
		 gpointer data)
{
	GdkGLContext* pGlContext = gtk_widget_get_gl_context (pWidget);
	GdkGLDrawable* pGlDrawable = gtk_widget_get_gl_drawable (pWidget);
	GLsizei iWidth = pWidget->allocation.width;
	GLsizei iHeight = pWidget->allocation.height;

	/* make GL-context "current" */
	if (!gdk_gl_drawable_gl_begin (pGlDrawable, pGlContext))
		return;

	opengl_init (iWidth, iHeight);

	/* end drawing to current GL-context */
	gdk_gl_drawable_gl_end (pGlDrawable);
}

gboolean
configure_handler (GtkWidget* pWidget,
		   GdkEventConfigure* pEvent,
		   gpointer data)
{
	GdkGLContext* pGlContext = gtk_widget_get_gl_context (pWidget);
	GdkGLDrawable* pGlDrawable = gtk_widget_get_gl_drawable (pWidget);
	GLsizei iWidth = pWidget->allocation.width;
	GLsizei iHeight = pWidget->allocation.height;

	/* make GL-context "current" */
	if (!gdk_gl_drawable_gl_begin (pGlDrawable, pGlContext))
		return FALSE;

	opengl_resize (iWidth, iHeight);

	/* end drawing to current GL-context */
	gdk_gl_drawable_gl_end (pGlDrawable);

	return TRUE;
}

gboolean
expose_handler (GtkWidget* pWidget,
		GdkEventExpose* pEvent,
		gpointer data)
{
	GdkGLContext* pGlContext = gtk_widget_get_gl_context (pWidget);
	GdkGLDrawable* pGlDrawable = gtk_widget_get_gl_drawable (pWidget);
	GLsizei iWidth = pWidget->allocation.width;
	GLsizei iHeight = pWidget->allocation.height;

	/* make GL-context "current" */
	if (!gdk_gl_drawable_gl_begin (pGlDrawable, pGlContext))
		return FALSE;

	opengl_draw (iWidth, iHeight, g_fFactor);

	if (gdk_gl_drawable_is_double_buffered (pGlDrawable))
		gdk_gl_drawable_swap_buffers (pGlDrawable);
	else
		glFlush ();

	/* end drawing to current GL-context */
	gdk_gl_drawable_gl_end (pGlDrawable);

	return TRUE;
}

gboolean
key_press_handler (GtkWidget* pWidget,
		   GdkEventKey* pEvent,
		   GMainLoop* pLoop)
{
	if (pEvent->type == GDK_KEY_PRESS)
	{
		switch (pEvent->keyval)
		{
			/* exit program */
			case GDK_q :
			case GDK_Escape :
				delete_handler (pWidget, NULL, NULL);
			break;

			/* toggle animation/refreshs, a bit awkward */
			case GDK_space :
				g_bDraw = g_bDraw ? FALSE : TRUE;
			break;

			default :
			break;
		}
	}

	return TRUE;
}

gboolean
scroll_handler (GtkWidget* pWidget,
		GdkEventScroll* pEvent,
		gpointer	data)
{
	switch (pEvent->direction)
	{
		case GDK_SCROLL_UP:
			if (g_fFactor < 25.0f)
				g_fFactor += 0.05f;
		break;

		case GDK_SCROLL_DOWN:
			if (g_fFactor > -25.0f)
				g_fFactor -= 0.05f;
		break;

		default:
		break;
	}

	g_print ("%f\n", g_fFactor);

	return TRUE;
}

gboolean
draw_handler (GtkWidget* pWidget)
{
	if (g_bDraw)
		gtk_widget_queue_draw (pWidget);

	return TRUE;
}
