/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** project:
**    OpenGL/cairo anti-aliasing trick
**
** file:
**    opengl.c
**
** author:
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**
** copyright (C) Mirco Mueller, December 2006, placed under the terms of the GPL
**
*******************************************************************************/

#include <gtk/gtk.h>
#include <gtk/gtkgl.h>
#include <gdk/gdkx.h>
#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/glu.h>
#include <GL/glx.h>

/*#include "shaders.h"*/
#include "cairo-utils.h"

#define Z_NEAR  1.0f
#define Z_FAR  50.0f
#define FOVY   20.0f

gboolean g_bHasShaders = FALSE;
GLuint g_auiFragProgs[6] = {0, 0, 0, 0, 0, 0};
Texture* g_apTextures[6];
Texture* g_apTextTextures[3];

/*typedef GLvoid
(*glActiveTextureFunc) (GLenum);

typedef GLvoid
(*glMultiTexCoord2fARBFunc) (GLenum, GLfloat, GLfloat);

glActiveTextureFunc glActiveTexture = 0;
glMultiTexCoord2fARBFunc glMultiTexCoord2fARB = 0;*/

typedef GLvoid
(*glXSwapIntervalSGIFunc) (GLint);

typedef GLvoid
(*glXSwapIntervalMESAFunc) (GLint);

/*void
opengl_enable_multitexture (void)
{
	const gchar* pcDummy = NULL;

	pcDummy = glXQueryExtensionsString (GDK_DISPLAY_XDISPLAY
					    (gdk_display_get_default ()),
					    0);
	if (g_strrstr (pcDummy, "GL_ARB_multitexture") != NULL)
	{
		glActiveTexture = (glActiveTextureFunc)
			gdk_gl_get_proc_address ("glActiveTexture");
		if (!glActiveTexture)
			g_print ("Could not get glActiveTexture()\n");

		glMultiTexCoord2fARB = (glMultiTexCoord2fARBFunc)
			gdk_gl_get_proc_address ("glMultiTexCoord2fARB");
		if (!glMultiTexCoord2fARB)
			g_print ("Could not get glMultiTexCoord2fARB()\n");
	}
}*/

GLuint
upload_program (GLenum iTarget,   /* either vertex- or fragment-target    */
		gchar* pcFilename /* full path/filename to shader-program */)
{
	GLuint uiProgramId = 0;
	gchar* pcBuffer = NULL;
	const gsize iBufferSize = 16 * 1024; /* FIXME */
	gsize iLength = 0;
	GError* pError = NULL;
	GLint iErrorPos = 0;
	GLint iIsNative = 0;

	/* make sure the filename is valid */
	g_assert (pcFilename != NULL);

	/* allocate new handle for shader-program */
	glGenProgramsARB (1, &uiProgramId);
	if (uiProgramId == 0)
	{
		g_print ("Could not generate shader-handle for \"%s\"!\n",
			 pcFilename);
		return 0;
	}

	/* allocate buffer for shader-source */
	pcBuffer = (gchar*) g_malloc0 (iBufferSize);
	if (!pcBuffer)
	{
		g_print ("Error allocating buffer for \"%s\" shader!\n",
			 pcFilename);
		glDeleteProgramsARB (1, &uiProgramId);
		return 0;
	}

	/* read shader-file into buffer */
	if (!g_file_get_contents (pcFilename,
				  &pcBuffer,
				  &iLength,
				  &pError))
	{
		g_print ("Could read shader-file \"%s\"!\n", pcFilename);
		g_free (pcBuffer);
		glDeleteProgramsARB (1, &uiProgramId);		
		return 0;
	}

	/* upload the shader-source to GL */
	glBindProgramARB (iTarget, uiProgramId);
	glProgramStringARB (iTarget,
			    GL_PROGRAM_FORMAT_ASCII_ARB,
			    iLength,
			    pcBuffer);

	/* free buffer for shader-source */
	g_free (pcBuffer);

	/* get any errors */
	glGetIntegerv (GL_PROGRAM_ERROR_POSITION_ARB, &iErrorPos);
	glGetProgramivARB (iTarget,
			   GL_PROGRAM_UNDER_NATIVE_LIMITS_ARB,
			   &iIsNative);

	/* check for errors */
	if ((iErrorPos == -1) && (iIsNative == 1))
		return uiProgramId;
	else
	{
		g_print ("Syntax/limit-error encountered in shader \"%s\"!\n\n",
			 pcFilename);
		g_print ("Found error:\n\033[31m%s\033[0m\n",
			 glGetString (GL_PROGRAM_ERROR_STRING_ARB));
		glDeleteProgramsARB (1, &uiProgramId);		
		return 0;
	}
}

/* enable (opengl_enable_vsync (2)) or disable (opengl_enable_vsync (0))
** the synchronization of the GL-rendering to the vblank of the display, for
** nice smooth graphics... you don't want to leave home without it;
** function-pointers are not saved as this call, because it is usually only
** needed once, thus it would not make much sense to keep the pointers around */
void
opengl_enable_vsync (gint iInterval)
{
	const gchar* pcDummy = NULL;
	glXSwapIntervalSGIFunc glXSwapIntervalSGI = 0;
	glXSwapIntervalMESAFunc glXSwapIntervalMESA = 0;

	pcDummy = glXQueryExtensionsString (GDK_DISPLAY_XDISPLAY
		(gdk_display_get_default ()), 0);

	if (g_strrstr (pcDummy, "GLX_SGI_swap_control") != NULL)
	{
		glXSwapIntervalSGI = (glXSwapIntervalSGIFunc)
			gdk_gl_get_proc_address ("glXSwapIntervalSGI");

		if (glXSwapIntervalSGI)
			glXSwapIntervalSGI (iInterval);
		else
			g_print ("Could not get glXSwapIntervalSGI()\n");
	}
	else if (g_strrstr (pcDummy, "GLX_MESA_swap_control") != NULL)
	{
		glXSwapIntervalMESA = (glXSwapIntervalMESAFunc)
			gdk_gl_get_proc_address ("glXSwapIntervalMESA");

		if (glXSwapIntervalMESA)
			glXSwapIntervalMESA (iInterval);
		else
			g_print ("Could not get glXSwapIntervalMESA()\n");
	}
	else
		g_print ("Unable to switch sync-to-vblank on.\n");
}

GLenum
opengl_what_texture_target (GLuint iWidth,
			    GLuint iHeight)
{
	GLenum iTarget = 0;
	gboolean bWidthIsPOT = FALSE;
	gboolean bHeightIsPOT = FALSE;
	gint i = 1;

	/* test width for power-of-two */
	for (i = 1; i <= 32; i++)
		if ((unsigned long) iWidth - (1L << i) == 0)
		{
			bWidthIsPOT = TRUE;
			break;
		}

	/* test height for power-of-two */
	for (i = 1; i <= 32; i++)
		if ((unsigned long) iHeight - (1L << i) == 0)
		{
			bHeightIsPOT = TRUE;
			break;
		}

	/* final conclusion */
	if (bWidthIsPOT && bHeightIsPOT)
		iTarget = GL_TEXTURE_2D;
	else
		iTarget = GL_TEXTURE_RECTANGLE_ARB;

	return iTarget;	
}

/* when you write GL-programs you will make mistakes, there's no doubt about 
** that; to make it a bit easier to identify a possible trouble-maker (e.g.
** something just not renders correctly or at all, but the program does not 
** crash, so gdb will not be of any help) you'll need something different, that
** is what opengl_print_error() is meant for; use it like this:
**
**   ...
**   glGetError ();
**   glTroubleMakerFunc ();
**   opengl_print_error ("glTroubleMakerFunc");
**   ...
**
** the first call clears GL's error-flag, then make your suspicious call and
** finally execute opengl_print_error() and see during runtime what gets spit
** out to the console; for the exact meaning of a reported error you have to
** consult the official GL-documentation (man-pages, red-book, orange-book) */
void
opengl_print_error (gchar* pcFuncName)
{
	GLenum error;

	error = glGetError ();

	switch (error)
	{
		case GL_NO_ERROR :
			g_print ("%s() caused GL_NO_ERROR\n", pcFuncName);
		break;

		case GL_INVALID_ENUM :
			g_print ("%s() caused GL_INVALID_ENUM\n", pcFuncName);
		break;

		case GL_INVALID_VALUE :
			g_print ("%s() GL_INVALID_VALUE\n", pcFuncName);
		break;

		case GL_INVALID_OPERATION :
			g_print ("%s() GL_INVALID_OPERATION\n", pcFuncName);
		break;

		case GL_STACK_OVERFLOW :
			g_print ("%s() GL_STACK_OVERFLOW\n", pcFuncName);
		break;

		case GL_STACK_UNDERFLOW :
			g_print ("%s() GL_STACK_UNDERFLOW\n", pcFuncName);
		break;

		case GL_OUT_OF_MEMORY :
			g_print ("%s() GL_OUT_OF_MEMORY\n", pcFuncName);
		break;

		case GL_TABLE_TOO_LARGE :
			g_print ("%s() GL_TABLE_TOO_LARGE\n", pcFuncName);
		break;
	}
}

/* easily switch between parallel and perspective projection */
void
opengl_change_projection (gboolean bParallel,/* paral. or persp. projection  */
			  GLsizei iWidth,    /* current width of GL-context  */
			  GLsizei iHeight    /* current height of GL-context */)
{
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	if (bParallel)
		glOrtho ((GLdouble) -iWidth / 2, (GLdouble) iWidth / 2,
			 (GLdouble) -iHeight / 2, (GLdouble) iHeight / 2,
			 Z_NEAR, Z_FAR);
	else
		gluPerspective (2.0f * FOVY,
				(GLdouble) iWidth / (GLdouble) iHeight,
				Z_NEAR,
				Z_FAR);

	glMatrixMode (GL_MODELVIEW);
}

/* GL_TEXTURE_3D, GL_PROXY_TEXTURE_CUBE_MAP and mip-mapping are not checked for
** in opengl_texture_proxy_check(), as they are of no interest in this
** application, they can of course be added later on, if the need arises */
gboolean
opengl_texture_proxy_check (GLenum iTarget, /* GL texture-target to test  */
			    gint iChannels, /* number of color-components */
			    gint iFormat,   /* GL image-format to test    */
			    gint iWidth,    /* width of texture to test   */
			    gint iHeight    /* height of texture to test  */)
{
	GLenum iProxyTarget = 0;
	GLint iProxyWidth = 0;
	GLint iProxyHeight = 0;

	/* select appropriate proxy-target */
	switch (iTarget)
	{
		case GL_TEXTURE_1D :
			iProxyTarget = GL_PROXY_TEXTURE_1D;
			glTexImage1D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;

		case GL_TEXTURE_2D :
			iProxyTarget = GL_PROXY_TEXTURE_2D;
			glTexImage2D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      iHeight,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;

		case GL_TEXTURE_RECTANGLE_ARB :
			iProxyTarget = GL_PROXY_TEXTURE_RECTANGLE_ARB;
			glTexImage2D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      iHeight,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;
	}

	/* query the results from the proxy-texture-target */
	glGetTexLevelParameteriv (iProxyTarget,
				  0,
				  GL_TEXTURE_WIDTH,
				  &iProxyWidth);
	glGetTexLevelParameteriv (iProxyTarget,
				  0,
				  GL_TEXTURE_HEIGHT,
				  &iProxyHeight);

	/* do the final check if texture would have fitted within the limits */
	if (iWidth == iProxyWidth &&
	    iHeight == iProxyHeight)
		return TRUE;
	else
		return FALSE;
}

/* very handy to "just quickly" load an image-file from disk and create a GL
** texture-object from it; note though that you have to know if the texture is
** pot or npot and set the iTarget (GL_TEXTURE_1D, GL_TEXTURE_2D... etc.)
** accordingly */
Texture*
opengl_upload_texture (GLuint	uiUnit,     /* GL multi-texture unit to use  */
		       gchar*	pcFilename, /* full path/filename of image   */
		       gboolean	bBlur,	    /* blur image or not via CPU     */
		       gboolean bGlow       /* create glow image via CPU     */)
{
	GLuint	   uiTextureId	   = 0;
	GdkPixbuf* pSrcPixbuf	   = NULL;
	GError*	   pError	   = NULL;
	gint	   iBitsPerChannel = 0;
	gint	   iWidth	   = 0;
	gint	   iHeight	   = 0;
	guchar*	   pucPixelBuffer  = NULL;
	gint	   iChannels	   = 0;
	guchar*	   pucGlowBuffer   = NULL;
	gint	   iGlowWidth	   = 0;
	gint	   iGlowHeight	   = 0;
	Texture*   pTexture	   = NULL;

	/* sanity check */
	g_assert (pcFilename != NULL);

	/* create a texture-object */
	pTexture = (Texture*) g_malloc0 (sizeof (Texture));
	if (!pTexture)
		return NULL;

	/* create a GL texture-handle */
	glGenTextures (1, &uiTextureId);
	if (uiTextureId == 0)
	{
		g_free ((gpointer) pTexture);
		return NULL;
	}

	/* load the image from disk */
	pSrcPixbuf = gdk_pixbuf_new_from_file (pcFilename, &pError);
	if (pSrcPixbuf == NULL)
	{
		g_free ((gpointer) pTexture);
		glDeleteTextures (1, &uiTextureId);
		return NULL;
	}

	/* only allow 8 bit per color-component */
	iBitsPerChannel = gdk_pixbuf_get_bits_per_sample (pSrcPixbuf);
	if (iBitsPerChannel != 8)
	{
		g_object_unref (pSrcPixbuf);
		g_free ((gpointer) pTexture);
		glDeleteTextures (1, &uiTextureId);
		return NULL;
	}

	/* determine the number of components */
	iChannels = gdk_pixbuf_get_n_channels (pSrcPixbuf);

	/* get hold of dimensions and pixel-data */
	iWidth = gdk_pixbuf_get_width (pSrcPixbuf);
	iHeight = gdk_pixbuf_get_height (pSrcPixbuf);
	pucPixelBuffer = gdk_pixbuf_get_pixels (pSrcPixbuf);

	if (bBlur && !bGlow)
		blur (pucPixelBuffer,
		      iWidth,
		      iHeight,
		      iChannels,
		      5);

	if (bGlow)
		glow (pucPixelBuffer,
		      iWidth,
		      iHeight,
		      iChannels,
      		      pucGlowBuffer,
		      &iGlowWidth,
		      &iGlowHeight);

	pTexture->uiTextureId	= uiTextureId;
	pTexture->uiUnit	= uiUnit;

	if (bGlow)
	{		
		pTexture->fWidth  = (GLfloat) iGlowWidth;
		pTexture->fHeight = (GLfloat) iGlowHeight;
	}
	else
	{
		pTexture->fWidth  = (GLfloat) iWidth;
		pTexture->fHeight = (GLfloat) iHeight;
	}

	pTexture->iTarget	= opengl_what_texture_target (pTexture->fWidth,
							      pTexture->fHeight);
	pTexture->iBufferSize	= 0;
	pTexture->pucBufferData	= NULL;

	pTexture->fMinS = 0.0f;
	pTexture->fMinT = 0.0f;
	pTexture->fMaxS = (pTexture->iTarget == GL_TEXTURE_2D) ? 1.0f : pTexture->fWidth;
	pTexture->fMaxT = (pTexture->iTarget == GL_TEXTURE_2D) ? 1.0f : pTexture->fHeight;

	/* see if the texture would fit */
	if (!opengl_texture_proxy_check (pTexture->iTarget,
					 iChannels,
					 (iChannels == 4) ? GL_RGBA : GL_RGB,
					 (GLint) pTexture->fWidth,
					 (GLint) pTexture->fHeight))
	{
		if (pucGlowBuffer)
			g_free ((gpointer) pucGlowBuffer);

		g_object_unref (pSrcPixbuf);
		g_free ((gpointer) pTexture);
		glDeleteTextures (1, &pTexture->uiTextureId);
		return NULL;
	}

	/* finally create the GL texture-object */
	glActiveTexture (pTexture->uiUnit);
	glBindTexture (pTexture->iTarget, pTexture->uiTextureId);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);
	if (bGlow)
		glTexImage2D (pTexture->iTarget,
			      0,
			      iChannels,
			      (GLsizei) pTexture->fWidth,
			      (GLsizei) pTexture->fHeight,
			      0,
			      (iChannels == 4) ? GL_RGBA : GL_RGB,
			      GL_UNSIGNED_BYTE,
			      pucGlowBuffer);
	else
		glTexImage2D (pTexture->iTarget,
			      0,
			      iChannels,
			      (GLsizei) pTexture->fWidth,
			      (GLsizei) pTexture->fHeight,
			      0,
			      (iChannels == 4) ? GL_RGBA : GL_RGB,
			      GL_UNSIGNED_BYTE,
			      pucPixelBuffer);

	g_object_unref (pSrcPixbuf);

	if (pucGlowBuffer)
		g_free ((gpointer) pucGlowBuffer);

	return pTexture;
}

Texture*
opengl_upload_mask (GLenum iTarget, /* GL texture-target to use     */
		    GLuint uiUnit,  /* GL multi-texture unit to use */
		    GLint iWidth,   /* width of mask texture        */
		    GLint iHeight   /* height of mask texture       */)
{
	GLuint	 uiTextureId	= 0;
	guchar*	 pucPixelBuffer	= NULL;
	Texture* pTexture	= NULL;
	gint	 i;

	/* create a texture-object */
	pTexture = (Texture*) g_malloc0 (sizeof (Texture));
	if (!pTexture)
		return NULL;

	/* create a GL texture-handle */
	glGenTextures (1, &uiTextureId);
	if (uiTextureId == 0)
	{
		g_free ((gpointer) pTexture);
		return NULL;
	}

	/* see if the texture would fit */
	if (!opengl_texture_proxy_check (iTarget,
					 4,
					 GL_RGBA,
					 iWidth,
					 iHeight))
	{
		g_free ((gpointer) pTexture);
		glDeleteTextures (1, &uiTextureId);
		return NULL;
	}

	/* create pixel-buffer */
	pucPixelBuffer = (guchar*) g_malloc0 (iWidth * iHeight * 4);
	if (!pucPixelBuffer)
	{
		g_free ((gpointer) pTexture);
		glDeleteTextures (1, &uiTextureId);
		return NULL;
	}

	for (i = 0; i < iWidth * iHeight * 4; i++)
		pucPixelBuffer[i] = 255;

	/* create mask */
	cairo_trim_image (pucPixelBuffer,
			  iWidth,
			  iHeight,
			  iWidth * 4);

	pTexture->iTarget	= iTarget;
	pTexture->uiTextureId	= uiTextureId;
	pTexture->uiUnit	= uiUnit;
	pTexture->fWidth	= (GLfloat) iWidth;
	pTexture->fHeight	= (GLfloat) iHeight;
	pTexture->iBufferSize	= 0;
	pTexture->pucBufferData	= NULL;

	/* finally create the GL texture-object */
	glActiveTexture (pTexture->uiUnit);
	glBindTexture (pTexture->iTarget, pTexture->uiTextureId);	
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);
	glTexImage2D (pTexture->iTarget,
		      0,
		      4,
		      (GLsizei) pTexture->fWidth,
		      (GLsizei) pTexture->fHeight,
		      0,
		      GL_RGBA,
		      GL_UNSIGNED_BYTE,
		      pucPixelBuffer);

	g_free ((gpointer) pucPixelBuffer);

	return pTexture;
}

Texture*
opengl_upload_text_texture (gchar* pcText, /* text to render into texture    */
			    GLuint uiUnit, /* GL multi-texture unit to use   */
			    GLint iWidth,  /* width of win. text appears in  */
			    GLint iHeight  /* height of win. text appears in */)
{
	GLuint uiTextureId = 0;
	Texture* pTexture = NULL;

	/* sanity check */
	g_assert (pcText != NULL);

	/* get new texture handle and verify its allocation */
	glGenTextures (1, &uiTextureId);
	if (uiTextureId == 0)
		return NULL;

	/* render text into a pixel-buffer via cairo */
	pTexture = cairo_text_to_texture (pcText);
	if (!pTexture)
	{
		glDeleteTextures (1, &uiTextureId);
		return NULL;
	}

	pTexture->uiTextureId = uiTextureId;
	pTexture->uiUnit = uiUnit;

	/* pass pixel-buffer to GL */
	glActiveTexture (pTexture->uiUnit);
	glBindTexture (pTexture->iTarget, pTexture->uiTextureId);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);
	glTexImage2D (pTexture->iTarget,
		      0,
		      GL_RGBA,
		      (GLsizei) pTexture->fWidth,
		      (GLsizei) pTexture->fHeight,
		      0,
		      GL_BGRA,
		      GL_UNSIGNED_BYTE,
		      pTexture->pucBufferData);

	/* get rid of resources no longer needed, buffer was allocated in
	** cairo_text_to_texture() */
	g_free ((gpointer) pTexture->pucBufferData);

	/* pass valid texture-handle to caller */
	return pTexture;
}

void
opengl_delete_texture (Texture* pTexture /* texture to delete */)
{
	if (!pTexture);
		return;

	glDeleteTextures (1, &pTexture->uiTextureId);

	if (pTexture->pucBufferData)
		g_free ((gpointer) pTexture->pucBufferData);

	g_free ((gpointer) pTexture);
}

/* just to make the background less boring I've implemented a simple
** checker-board painting function; remember: values to pass differ greatly
** depending on the kind of projection-matrix you've set up */
void
opengl_paint_checker (GLfloat fStartX, /* x-coord of lower-left corner */
		      GLfloat fStartY, /* y-coord of lower-left corner */
		      GLfloat fSize,   /* width/height of one tile     */
		      GLint iColumns,  /* number of columns to draw    */
		      GLint iRows      /* number of rows to draw       */)
{
	GLfloat fX = fStartX;
	GLfloat fY = fStartY;
	gboolean bColorToggle = TRUE;
	gboolean bStartedWith = TRUE;
	gint x;
	gint y;

	glBegin (GL_QUADS);
	for (x = 0; x < iColumns; x++)
	{
		bColorToggle = bStartedWith ? FALSE : TRUE;
		bStartedWith = bColorToggle;

		for (y = 0; y < iRows; y++)
		{
			if (bColorToggle)
				glColor4f (0.4f, 0.4f, 0.4f, 1.0f);
			else
				glColor4f (0.5f, 0.5f, 0.5f, 1.0f);

			bColorToggle = bColorToggle ? FALSE : TRUE;

			glVertex3f (fX,         fY,         0.0f);
			glVertex3f (fX,         fY + fSize, 0.0f);
			glVertex3f (fX + fSize, fY + fSize, 0.0f);
			glVertex3f (fX + fSize, fY,         0.0f);

			fY += fSize;
		}
		fX += fSize;
		fY = fStartY;
	}
	glEnd ();
}

/* this paints a nice white-greyish gradient; meant to be used in parallel
** projection */
void
opengl_paint_bg (GLfloat fStartX, /* x-coord of lower-left corner */
		 GLfloat fStartY, /* y-coord of lower-left corner */
		 GLfloat fWidth,  /* width of background          */
		 GLfloat fHeight  /* height of background         */)
{
	glBegin (GL_QUADS);

	/* top half, black */
	glColor3f (0.0f, 0.0f, 0.0f);
	glVertex3f (fStartX,          fStartY + fHeight,        0.0f);
	glVertex3f (fStartX + fWidth, fStartY + fHeight,        0.0f);
	glVertex3f (fStartX + fWidth, fStartY + fHeight / 2.0f, 0.0f);
	glVertex3f (fStartX,          fStartY + fHeight / 2.0f, 0.0f);

	/* bottom half, black to dark grey */
	glColor3f (0.0f, 0.0f, 0.0f);
	glVertex3f (fStartX + fWidth, fStartY + fHeight / 2.0f, 0.0f);
	glVertex3f (fStartX,          fStartY + fHeight / 2.0f, 0.0f);
	glColor3f (0.3f, 0.3f, 0.3f);
	glVertex3f (fStartX,          fStartY,                  0.0f);
	glVertex3f (fStartX + fWidth, fStartY,                  0.0f);

	glEnd ();
}

/* the initial setup for GL, setting some required states, position the camera,
** setup shaders, switch on vsync, load/create textures */
void
opengl_init (GLint iWidth,
	     GLint iHeight)
{
	GLfloat afLightDiffuse[4] = {0.76f, 0.75f, 0.65f, 1.0f};

	/* set some GL-states */
	glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
	glColor4f (1.0f, 1.0f, 1.0f, 0.0f);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, 1);

	/* "switch on the light" */
	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, afLightDiffuse);

	/* setup "camera" */
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	gluLookAt (0.0f, 0.0f, Z_NEAR,
		   0.0f, 0.0f, 0.0f,
		   0.0f, 1.0f, 0.0f);
	glTranslatef (0.0f, 0.0f, -Z_NEAR);

	/* try to find and initialize programmable shading capabilities */
	/*if (setup_shader_callbacks ())
	{*/
		g_bHasShaders = TRUE;
		glEnable (GL_FRAGMENT_PROGRAM_ARB);
		g_auiFragProgs[0] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "solid.frag");
		g_auiFragProgs[1] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "texture-2d.frag");
		g_auiFragProgs[2] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "texture-rect.frag");
		g_auiFragProgs[3] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "texture-blur-2d.frag");
		g_auiFragProgs[4] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "texture-emboss-2d.frag");
		g_auiFragProgs[5] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "texture-sobel-2d.frag");
	/*}*/

	/* setup multi-texture related calls */
	/*opengl_enable_multitexture ();*/

	/* switch on sync-to-vblank*/
	opengl_enable_vsync (1);

	/* load and setup needed texture-resources */
	g_apTextures[0] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	g_apTextures[1] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	g_apTextures[2] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	g_apTextures[3] = opengl_upload_mask (GL_TEXTURE_2D,
					      GL_TEXTURE1,
					      512,
					      512);
	g_apTextures[4] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	g_apTextures[5] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);

	g_apTextTextures[0] =  opengl_upload_text_texture ("embossed (GPU)",
							   GL_TEXTURE0,
							   iWidth,
							   iHeight);
	g_apTextTextures[1] =  opengl_upload_text_texture ("blurred (GPU)",
							   GL_TEXTURE0,
							   iWidth,
							   iHeight);
	g_apTextTextures[2] =  opengl_upload_text_texture ("edge-detect (GPU)",
							   GL_TEXTURE0,
							   iWidth,
							   iHeight);
}

/* needed to ensure proper resize-behaviour of the GL-context if the window is
** altered in size by the user */
void
opengl_resize (GLsizei iWidth, /* new width to adjust the GL-context to  */
	       GLsizei iHeight /* new height to adjust the GL-context to */)
{
	glViewport (0, 0, iWidth, iHeight);
	opengl_change_projection (TRUE, iWidth, iHeight);
	glClear (GL_COLOR_BUFFER_BIT);

	/* for the extreme rare case someone might try to compile this under
	** windows (you'll have to do some work on opengl_enable_vsync() then)
	** I've added the following calls; under Linux/Unix they are generally
	** not needed, but under Windows a resize of the GL-context invalidates
	** any texture-object, thus it has to be recreated (a little sidenote
	** regarding this issue: remember id-software games, which are OpenGL-
	** -based... they regenerate all their game resources if you alter the 
	** game-resolution) */
	opengl_delete_texture (g_apTextures[0]);
	opengl_delete_texture (g_apTextures[1]);
	opengl_delete_texture (g_apTextures[2]);
	opengl_delete_texture (g_apTextures[3]);
	opengl_delete_texture (g_apTextures[4]);
	g_apTextures[0] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	g_apTextures[1] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	g_apTextures[2] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	g_apTextures[3] = opengl_upload_mask (GL_TEXTURE_2D,
					      GL_TEXTURE1,
					      512,
					      512);
	g_apTextures[4] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	g_apTextures[5] = opengl_upload_texture (GL_TEXTURE0,
						 "ui-shot.png",
						 FALSE,
						 FALSE);
	
	opengl_delete_texture (g_apTextTextures[0]);
	opengl_delete_texture (g_apTextTextures[1]);
	opengl_delete_texture (g_apTextTextures[2]);
	g_apTextTextures[0] =  opengl_upload_text_texture ("embossed (GPU)",
							   GL_TEXTURE0,
							   iWidth,
							   iHeight);
	g_apTextTextures[1] =  opengl_upload_text_texture ("blurred (GPU)",
							   GL_TEXTURE0,
							   iWidth,
							   iHeight);
	g_apTextTextures[2] =  opengl_upload_text_texture ("edge-detect (GPU)",
							   GL_TEXTURE0,
							   iWidth,
							   iHeight);
}

void
opengl_draw_textured_quad (Texture* pTexture,   /* texture to use          */
			   GLuint   uiFragProg, /* shader to use           */
			   GLfloat* pfPosition, /* position of quad        */
			   GLfloat  fAngle,     /* current rotation angle  */
			   GLfloat* pfAxis,     /* rotation-axis           */
			   gboolean bNormal,    /* normal or reflected     */
			   GLfloat  fValue      /* value to pass to shader */)
{
	GLfloat afMaterialDiffuseFrontFull[4]    = {1.0f, 1.0f, 1.0f, 1.0f};
	GLfloat afMaterialDiffuseBackFull[4]     = {0.0f, 0.16f, 0.31f, 1.0f};
	GLfloat afMaterialDiffuseFrontOff[4]     = {1.0f, 1.0f, 1.0f, 0.0f};
	GLfloat afMaterialDiffuseBackOff[4]      = {0.0f, 0.16f, 0.31f, 0.0f};
	GLfloat afMaterialDiffuseFrontQuarter[4] = {1.0f, 1.0f, 1.0f, 0.25f};
	GLfloat afMaterialDiffuseBackQuarter[4]  = {0.0f, 0.16f, 0.31f, 0.25f};
	GLfloat fPixelSize = 0.0f;

	/* sanity checks */
	g_assert (pTexture != NULL);
	g_assert (pfPosition != NULL);
	g_assert (pfAxis != NULL);

	glActiveTexture (pTexture->uiUnit);
	glEnable (pTexture->iTarget);
	glBindTexture (pTexture->iTarget, pTexture->uiTextureId);
	glBindProgramARB (GL_FRAGMENT_PROGRAM_ARB, uiFragProg);

	fPixelSize = 1.0f / pTexture->fWidth;

	glProgramLocalParameter4fARB (GL_FRAGMENT_PROGRAM_ARB,
				      0,
				      fValue,
				      -fValue,
				      fPixelSize,
				      0.0f);

	glPushMatrix ();
	glTranslatef (pfPosition[X], pfPosition[Y], pfPosition[Z]);
	glRotatef (fAngle, pfAxis[X], pfAxis[Y], pfAxis[Z]);

	glBegin (GL_QUADS);

	if (bNormal)
	{
		glNormal3f (0.0f, 0.0f, 1.0f);
		glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
		glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontFull);
		glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackFull);
		glMultiTexCoord2fARB (pTexture->uiUnit,
				      pTexture->fMinS,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE2,
				      pTexture->fMinS,
				      pTexture->fMaxT - fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE3,
				      pTexture->fMinS + fPixelSize,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE4,
				      pTexture->fMinS,
				      pTexture->fMaxT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE5,
				      pTexture->fMinS - fPixelSize,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE6,
				      pTexture->fMinS,
				      pTexture->fMaxT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE7,
				      pTexture->fMinS + fPixelSize,
				      pTexture->fMaxT);
		glVertex3f (-1.5f, -1.525f, 0.0f);

		glMultiTexCoord2fARB (pTexture->uiUnit,
				      pTexture->fMaxS,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE2,
				      pTexture->fMaxS,
				      pTexture->fMaxT - fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE3,
				      pTexture->fMaxS + fPixelSize,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE4,
				      pTexture->fMaxS,
				      pTexture->fMaxT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE5,
				      pTexture->fMaxS - fPixelSize,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE6,
				      pTexture->fMaxS,
				      pTexture->fMaxT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE7,
				      pTexture->fMaxS + fPixelSize,
				      pTexture->fMaxT);
		glVertex3f ( 1.5f, -1.525f, 0.0f);

		glMultiTexCoord2fARB (pTexture->uiUnit,
				      pTexture->fMaxS,
				      pTexture->fMinT);
		glMultiTexCoord2fARB (GL_TEXTURE2,
				      pTexture->fMaxS,
				      pTexture->fMinT - fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE3,
				      pTexture->fMaxS + fPixelSize,
				      pTexture->fMinT);
		glMultiTexCoord2fARB (GL_TEXTURE4,
				      pTexture->fMaxS,
				      pTexture->fMinT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE5,
				      pTexture->fMaxS - fPixelSize,
				      pTexture->fMinT);
		glMultiTexCoord2fARB (GL_TEXTURE6,
				      pTexture->fMaxS,
				      pTexture->fMinT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE7,
				      pTexture->fMaxS + fPixelSize,
				      pTexture->fMinT);
		glVertex3f ( 1.5f,  1.475f, 0.0f);

		glMultiTexCoord2fARB (pTexture->uiUnit,
				      pTexture->fMinS,
				      pTexture->fMinT);
		glMultiTexCoord2fARB (GL_TEXTURE2,
				      pTexture->fMinS,
				      pTexture->fMinT - fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE3,
				      pTexture->fMinS + fPixelSize,
				      pTexture->fMinT);
		glMultiTexCoord2fARB (GL_TEXTURE4,
				      pTexture->fMinS,
				      pTexture->fMinT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE5,
				      pTexture->fMinS - fPixelSize,
				      pTexture->fMinT);
		glMultiTexCoord2fARB (GL_TEXTURE6,
				      pTexture->fMinS,
				      pTexture->fMinT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE7,
				      pTexture->fMinS + fPixelSize,
				      pTexture->fMinT);
		glVertex3f (-1.5f,  1.475f, 0.0f);
	}
	else
	{
		glNormal3f (0.0f, 0.0f, 1.0f);
		glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
		glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontOff);
		glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackOff);
		glMultiTexCoord2fARB (pTexture->uiUnit,
				      pTexture->fMinS,
				      pTexture->fMaxT * 0.5f);
		glMultiTexCoord2fARB (GL_TEXTURE2,
				      pTexture->fMinS,
				      pTexture->fMaxT * 0.5f - fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE3,
				      pTexture->fMinS + fPixelSize,
				      pTexture->fMaxT * 0.5f);
		glMultiTexCoord2fARB (GL_TEXTURE4,
				      pTexture->fMinS,
				      pTexture->fMaxT * 0.5f + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE5,
				      pTexture->fMinS - fPixelSize,
				      pTexture->fMaxT * 0.5f);
		glMultiTexCoord2fARB (GL_TEXTURE6,
				      pTexture->fMinS,
				      pTexture->fMaxT * 0.5f + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE7,
				      pTexture->fMinS + fPixelSize,
				      pTexture->fMaxT * 0.5f);
		glVertex3f (-1.5f, -2.75f, 0.0f);

		glMultiTexCoord2fARB (pTexture->uiUnit,
				      pTexture->fMaxS,
				      pTexture->fMaxT * 0.5f);
		glMultiTexCoord2fARB (GL_TEXTURE2,
				      pTexture->fMaxS,
				      pTexture->fMaxT * 0.5f - fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE3,
				      pTexture->fMaxS + fPixelSize,
				      pTexture->fMaxT * 0.5f);
		glMultiTexCoord2fARB (GL_TEXTURE4,
				      pTexture->fMaxS,
				      pTexture->fMaxT * 0.5f + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE5,
				      pTexture->fMaxS - fPixelSize,
				      pTexture->fMaxT * 0.5f);
		glMultiTexCoord2fARB (GL_TEXTURE6,
				      pTexture->fMaxS,
				      pTexture->fMaxT * 0.5f + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE7,
				      pTexture->fMaxS + fPixelSize,
				      pTexture->fMaxT * 0.5f);
		glVertex3f ( 1.5f, -2.75f, 0.0f);

		glColor4f (1.0f, 1.0f, 1.0f, 0.25f);
		glMaterialfv (GL_FRONT,
			      GL_DIFFUSE,
			      afMaterialDiffuseFrontQuarter);
		glMaterialfv (GL_BACK,
			      GL_DIFFUSE,
			      afMaterialDiffuseBackQuarter);
		glMultiTexCoord2fARB (pTexture->uiUnit,
				      pTexture->fMaxS,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE2,
				      pTexture->fMaxS,
				      pTexture->fMaxT - fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE3,
				      pTexture->fMaxS + fPixelSize,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE4,
				      pTexture->fMaxS,
				      pTexture->fMaxT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE5,
				      pTexture->fMaxS - fPixelSize,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE6,
				      pTexture->fMaxS,
				      pTexture->fMaxT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE7,
				      pTexture->fMaxS + fPixelSize,
				      pTexture->fMaxT);
		glVertex3f ( 1.5f, -1.475f, 0.0f);

		glMultiTexCoord2fARB (pTexture->uiUnit,
				      pTexture->fMinS,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE2,
				      pTexture->fMinS,
				      pTexture->fMaxT - fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE3,
				      pTexture->fMinS + fPixelSize,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE4,
				      pTexture->fMinS,
				      pTexture->fMaxT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE5,
				      pTexture->fMinS - fPixelSize,
				      pTexture->fMaxT);
		glMultiTexCoord2fARB (GL_TEXTURE6,
				      pTexture->fMinS,
				      pTexture->fMaxT + fPixelSize);
		glMultiTexCoord2fARB (GL_TEXTURE7,
				      pTexture->fMinS + fPixelSize,
				      pTexture->fMaxT);
		glVertex3f (-1.5f, -1.475f, 0.0f);
	}

	glEnd ();
	glPopMatrix ();
	glDisable (pTexture->iTarget);
}

void
opengl_blur_pass (Texture*	pSrcTexture,	/* texture to read from      */
		  Texture*	pDstTexture,	/* texture to write to       */
		  GLuint	uiFragProg,	/* shader to use             */
		  GLfloat	fDeviation,	/* std. deviatino to use     */
		  PassDirection	direction,	/* do which pass of the blur */
		  GLint		iWindowWidth,	/* width of window/context   */
		  GLint		iWindowHeight	/* height of window/context  */)
{
	static GLfloat fPixelSize = 0.0f;
	static GLfloat afKernel[12];
	static GLfloat aafVertices[4][3];

	opengl_change_projection (TRUE, iWindowWidth, iWindowHeight);
	glDisable (GL_LIGHTING);
	glEnable (pSrcTexture->iTarget);
	glBindProgramARB (GL_FRAGMENT_PROGRAM_ARB, uiFragProg);
	glActiveTexture (pSrcTexture->uiUnit);
	glBindTexture (pSrcTexture->iTarget,
		       pSrcTexture->uiTextureId);

	compute_1d_kernel (5, fDeviation, afKernel);

	if (direction == VERTICAL_PASS)
	{
		fPixelSize = 1.0f / pSrcTexture->fHeight;

		aafVertices[0][X] = 0.0f;
		aafVertices[0][Y] = pSrcTexture->fHeight;
		aafVertices[0][Z] = 0.0f;

		aafVertices[1][X] = pSrcTexture->fWidth;
		aafVertices[1][Y] = pSrcTexture->fHeight;
		aafVertices[1][Z] = 0.0f;

		aafVertices[2][X] = pSrcTexture->fWidth;
		aafVertices[2][Y] = 0.0f;
		aafVertices[2][Z] = 0.0f;

		aafVertices[3][X] = 0.0f;
		aafVertices[3][Y] = 0.0f;
		aafVertices[3][Z] = 0.0f;
	}
	else if (direction == HORIZONTAL_PASS)
	{
		fPixelSize = 1.0f / pSrcTexture->fWidth;

		aafVertices[0][X] = 0.0f;
		aafVertices[0][Y] = pSrcTexture->fHeight;
		aafVertices[0][Z] = 0.0f;

		aafVertices[1][X] = pSrcTexture->fWidth;
		aafVertices[1][Y] = pSrcTexture->fHeight;
		aafVertices[1][Z] = 0.0f;

		aafVertices[2][X] = pSrcTexture->fWidth;
		aafVertices[2][Y] = 0.0f;
		aafVertices[2][Z] = 0.0f;

		aafVertices[3][X] = 0.0f;
		aafVertices[3][Y] = 0.0f;
		aafVertices[3][Z] = 0.0f;
	}

	afKernel[11] = fPixelSize;

	glProgramLocalParameter4fARB (GL_FRAGMENT_PROGRAM_ARB,
				      0,
				      afKernel[0],
				      afKernel[1],
				      afKernel[2],
				      afKernel[3]);
	glProgramLocalParameter4fARB (GL_FRAGMENT_PROGRAM_ARB,
				      1,
				      afKernel[4],
				      afKernel[5],
				      afKernel[6],
				      afKernel[7]);
	glProgramLocalParameter4fARB (GL_FRAGMENT_PROGRAM_ARB,
				      2,
				      afKernel[8],
				      afKernel[9],
				      afKernel[10],
				      afKernel[11]);

	glPushMatrix ();
	glTranslatef ((GLfloat) iWindowWidth / -2.0f,
		      (GLfloat) iWindowHeight / -2.0f,
		      0.0f);
	glBegin (GL_QUADS);
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);

	glMultiTexCoord2fARB (GL_TEXTURE0, fPixelSize * -5.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE1, fPixelSize * -4.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE2, fPixelSize * -3.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE3, fPixelSize * -2.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE4, fPixelSize * -1.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE5, fPixelSize *  0.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE6, fPixelSize *  1.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE7, fPixelSize *  2.0f, 1.0f);
	glVertex3f (aafVertices[0][X], aafVertices[0][Y], aafVertices[0][Z]);

	glMultiTexCoord2fARB (GL_TEXTURE0, fPixelSize * -5.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE1, fPixelSize * -4.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE2, fPixelSize * -3.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE3, fPixelSize * -2.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE4, fPixelSize * -1.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE5, fPixelSize *  0.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE6, fPixelSize *  1.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE7, fPixelSize *  2.0f, 0.0f);
	glVertex3f (aafVertices[1][X], aafVertices[1][Y], aafVertices[1][Z]);

	glMultiTexCoord2fARB (GL_TEXTURE0, 1.0f + fPixelSize * -5.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE1, 1.0f + fPixelSize * -4.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE2, 1.0f + fPixelSize * -3.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE3, 1.0f + fPixelSize * -2.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE4, 1.0f + fPixelSize * -1.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE5, 1.0f + fPixelSize *  0.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE6, 1.0f + fPixelSize *  1.0f, 0.0f);
	glMultiTexCoord2fARB (GL_TEXTURE7, 1.0f + fPixelSize *  2.0f, 0.0f);
	glVertex3f (aafVertices[2][X], aafVertices[2][Y], aafVertices[2][Z]);

	glMultiTexCoord2fARB (GL_TEXTURE0, 1.0f + fPixelSize * -5.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE1, 1.0f + fPixelSize * -4.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE2, 1.0f + fPixelSize * -3.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE3, 1.0f + fPixelSize *  2.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE4, 1.0f + fPixelSize *  1.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE5, 1.0f + fPixelSize *  0.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE6, 1.0f + fPixelSize *  1.0f, 1.0f);
	glMultiTexCoord2fARB (GL_TEXTURE7, 1.0f + fPixelSize *  2.0f, 1.0f);
	glVertex3f (aafVertices[3][X], aafVertices[3][Y], aafVertices[3][Z]);

	glEnd ();
	glPopMatrix ();
	glDisable (pSrcTexture->iTarget);

	/* read-back the blurred image */
	glEnable (pDstTexture->iTarget);
	glBindTexture (pDstTexture->iTarget,
		       pDstTexture->uiTextureId);
	glCopyTexSubImage2D (pDstTexture->iTarget,
			     0,
			     0,
			     0,
			     0,
			     0,
		 	     pDstTexture->fWidth,
		 	     pDstTexture->fHeight);
	glDisable (pDstTexture->iTarget);
}

/* the main action takes place here once everything is setup; this is what gets
** called very time we force a refresh/redraw of the window */
void
opengl_draw (GLsizei iWidth,  /* current width of the GL-context  */
	     GLsizei iHeight, /* current height of the GL-context */
	     GLfloat fFactor  /* some factor to use for anything  */)
{
	static GLfloat fAngle = 0.0f;
	static GLfloat afAxis[] = {0.0f, 1.0f, 0.0f};
	static GLfloat afPosition[] = {0.0f, 0.0f, 0.0f};
	static GLfloat fOldFactor = 0.0f;

	glClear (GL_COLOR_BUFFER_BIT);

	if (fOldFactor != fFactor)
	{
		int iPass;

		opengl_blur_pass (g_apTextures[1],
				  g_apTextures[4],
				  g_auiFragProgs[3],
				  fFactor,
				  HORIZONTAL_PASS,
				  iWidth,
				  iHeight);
		opengl_blur_pass (g_apTextures[4],
				  g_apTextures[5],
				  g_auiFragProgs[3],
				  fFactor,
				  VERTICAL_PASS,
				  iWidth,
				  iHeight);
		for (iPass = 0; iPass < 5; iPass++)
		{
			opengl_blur_pass (g_apTextures[5],
					  g_apTextures[4],
					  g_auiFragProgs[3],
					  fFactor,
					  HORIZONTAL_PASS,
					  iWidth,
					  iHeight);
			opengl_blur_pass (g_apTextures[4],
					  g_apTextures[5],
					  g_auiFragProgs[3],
					  fFactor,
					  VERTICAL_PASS,
					  iWidth,
					  iHeight);
		}

		fOldFactor = fFactor;
	}

	/* disable texturing and lighting and draw the background using the
	** first shader */
	opengl_change_projection (TRUE, iWidth, iHeight);
	glDisable (GL_TEXTURE_RECTANGLE_ARB);
	glDisable (GL_TEXTURE_2D);
	glDisable (GL_LIGHTING);
	glBindProgramARB (GL_FRAGMENT_PROGRAM_ARB, g_auiFragProgs[0]);
	opengl_paint_bg (-iWidth/2.0f, -iHeight/2.0f, iWidth, iHeight);

	/* enable texturing and lighting and draw the spinning images with
	** other shaders */
	opengl_change_projection (FALSE, iWidth, iHeight);
	glEnable (GL_LIGHTING);
	fAngle += 1.5f;

	/* left thing */
	afPosition[X] = -3.5f;
	afPosition[Y] =  0.0f;
	afPosition[Z] = -5.0f;
	opengl_draw_textured_quad (g_apTextures[0],
				   g_auiFragProgs[4],
				   afPosition,
				   fAngle,
				   afAxis,
				   TRUE,
				   fFactor);
	opengl_draw_textured_quad (g_apTextures[0],
				   g_auiFragProgs[4],
				   afPosition,
				   fAngle,
				   afAxis,
				   FALSE,
				   fFactor);

	/* middle thing */
	afPosition[X] =  0.0f;
	afPosition[Y] =  0.0f;
	afPosition[Z] = -5.0f;
	opengl_draw_textured_quad (g_apTextures[5],
				   g_auiFragProgs[1],
				   afPosition,
				   0.0f,
				   afAxis,
				   TRUE,
				   fFactor);
	opengl_draw_textured_quad (g_apTextures[5],
				   g_auiFragProgs[1],
				   afPosition,
				   0.0f,
				   afAxis,
				   FALSE,
				   fFactor);

	/* right thing */
	afPosition[X] =  3.5f;
	afPosition[Y] =  0.0f;
	afPosition[Z] = -5.0f;
	opengl_draw_textured_quad (g_apTextures[2],
				   g_auiFragProgs[5],
				   afPosition,
				   fAngle,
				   afAxis,
				   TRUE,
				   fFactor);
	opengl_draw_textured_quad (g_apTextures[2],
				   g_auiFragProgs[5],
				   afPosition,
				   fAngle,
				   afAxis,
				   FALSE,
				   fFactor);

	/* draw the text-labels */
	opengl_change_projection (TRUE, iWidth, iHeight);
	glDisable (GL_TEXTURE_2D);
	glDisable (GL_LIGHTING);
	glBindProgramARB (GL_FRAGMENT_PROGRAM_ARB, g_auiFragProgs[2]);

	glActiveTexture (g_apTextTextures[0]->uiUnit);
	glEnable (g_apTextTextures[0]->iTarget);
	glBindTexture (g_apTextTextures[0]->iTarget,
		       g_apTextTextures[0]->uiTextureId);
	glPushMatrix ();
	glTranslatef (- iWidth / 3 - g_apTextTextures[0]->fWidth / 2.0f, -1.0f * (GLfloat) iHeight / 2 + 20.0f, 0.0f);
	glBegin (GL_QUADS);
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glMultiTexCoord2fARB (g_apTextTextures[0]->uiUnit,
			      0.0f,
			      g_apTextTextures[0]->fHeight);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[0]->uiUnit,
			      g_apTextTextures[0]->fWidth,
			      g_apTextTextures[0]->fHeight);
	glVertex3f (g_apTextTextures[0]->fWidth, 0.0f, 0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[0]->uiUnit,
			      g_apTextTextures[0]->fWidth,
			      0.0f);
	glVertex3f (g_apTextTextures[0]->fWidth,
		    g_apTextTextures[0]->fHeight,
		    0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[0]->uiUnit, 0.0f, 0.0f);
	glVertex3f (0.0f, g_apTextTextures[0]->fHeight, 0.0f);
	glEnd ();
	glPopMatrix ();
	glDisable (g_apTextTextures[0]->iTarget);

	glActiveTexture (g_apTextTextures[1]->uiUnit);
	glEnable (g_apTextTextures[1]->iTarget);
	glBindTexture (g_apTextTextures[1]->iTarget,
		       g_apTextTextures[1]->uiTextureId);
	glPushMatrix ();
	glTranslatef (g_apTextTextures[1]->fWidth / -2.0f, -1.0f * (GLfloat) iHeight / 2 + 20.0f, 0.0f);
	glBegin (GL_QUADS);
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glMultiTexCoord2fARB (g_apTextTextures[1]->uiUnit,
			      0.0f,
			      g_apTextTextures[1]->fHeight);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[1]->uiUnit,
			      g_apTextTextures[1]->fWidth,
			      g_apTextTextures[1]->fHeight);
	glVertex3f (g_apTextTextures[1]->fWidth, 0.0f, 0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[1]->uiUnit,
			      g_apTextTextures[1]->fWidth,
			      0.0f);
	glVertex3f (g_apTextTextures[1]->fWidth,
		    g_apTextTextures[1]->fHeight,
		    0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[1]->uiUnit,
			      0.0f,
			      0.0f);
	glVertex3f (0.0f, g_apTextTextures[1]->fHeight, 0.0f);
	glEnd ();
	glPopMatrix ();
	glDisable (g_apTextTextures[1]->iTarget);

	glActiveTexture (g_apTextTextures[2]->uiUnit);
	glEnable (g_apTextTextures[2]->iTarget);
	glBindTexture (g_apTextTextures[2]->iTarget,
		       g_apTextTextures[2]->uiTextureId);
	glPushMatrix ();
	glTranslatef (iWidth / 3 - g_apTextTextures[2]->fWidth / 2.0f, -1.0f * (GLfloat) iHeight / 2 + 20.0f, 0.0f);
	glBegin (GL_QUADS);
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glMultiTexCoord2fARB (g_apTextTextures[2]->uiUnit,
			      0.0f,
			      g_apTextTextures[2]->fHeight);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[2]->uiUnit,
			      g_apTextTextures[2]->fWidth,
			      g_apTextTextures[2]->fHeight);
	glVertex3f (g_apTextTextures[2]->fWidth, 0.0f, 0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[2]->uiUnit,
			      g_apTextTextures[2]->fWidth,
			      0.0f);
	glVertex3f (g_apTextTextures[2]->fWidth,
		    g_apTextTextures[2]->fHeight,
		    0.0f);
	glMultiTexCoord2fARB (g_apTextTextures[2]->uiUnit, 0.0f, 0.0f);
	glVertex3f (0.0f, g_apTextTextures[2]->fHeight, 0.0f);
	glEnd ();
	glPopMatrix ();
	glDisable (g_apTextTextures[2]->iTarget);
}

/* just a tiny bit of resource cleaning needed to happen at the end */
void
opengl_cleanup (void)
{
	if (g_bHasShaders)
		glDeleteProgramsARB (5, g_auiFragProgs);

	opengl_delete_texture (g_apTextures[0]);
	opengl_delete_texture (g_apTextures[1]);
	opengl_delete_texture (g_apTextures[2]);
	opengl_delete_texture (g_apTextures[3]);
	opengl_delete_texture (g_apTextures[4]);
	opengl_delete_texture (g_apTextures[5]);

	opengl_delete_texture (g_apTextTextures[0]);
	opengl_delete_texture (g_apTextTextures[1]);
	opengl_delete_texture (g_apTextTextures[2]);
}
