/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** project:
**    OpenGL/cairo anti-aliasing trick
**
** file:
**    opengl.h
**
** author:
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**
** copyright (C) Mirco Mueller, December 2006, placed under the terms of the GPL
**
*******************************************************************************/

#ifndef _OPENGL_H
#define _OPENGL_H

#include <gtk/gtk.h>
#include <GL/gl.h>

typedef enum
{
	R = 0,
	G,
	B,
	A
} ColorComponents;

typedef enum
{
	X = 0,
	Y,
	Z,
	W
} Coordinates;

typedef enum
{
	VERTICAL_PASS = 0,
	HORIZONTAL_PASS
} PassDirection;

typedef struct _Texture
{
	GLenum   iTarget;
	GLuint	 uiTextureId;
	GLuint	 uiUnit;
	GLfloat  fWidth;
	GLfloat  fHeight;
	GLfloat	 fMinS;
	GLfloat	 fMinT;
	GLfloat	 fMaxS;
	GLfloat	 fMaxT;
	GLuint   iBufferSize;
	GLubyte* pucBufferData; /* free as soon as no longer needed */
} Texture;

GLenum
opengl_what_texture_target (GLuint iWidth,
			    GLuint iHeight);

void
opengl_print_error (gchar* pcFuncName);

void
opengl_enable_vsync (gint iInterval);

void
opengl_change_projection (gboolean bParallel,/* paral. or persp. projection */
			  GLsizei iWidth,    /* current width of GL-context */
			  GLsizei iHeight    /* current height of GL-context*/);

gboolean
opengl_texture_proxy_check (GLenum iTarget, /* GL texture-target to test  */
			    gint iChannels, /* number of color-components */
			    gint iFormat,   /* GL image-format to test    */
			    gint iWidth,    /* width of texture to test   */
			    gint iHeight    /* height of texture to test  */);

Texture*
opengl_upload_texture (GLuint	uiUnit,	    /* GL multi-texture unit to use */
		       gchar*	pcFilename, /* full path/filename of image  */
		       gboolean	bBlur,	    /* blur image or not via CPU    */
		       gboolean	bGlow	    /* create glow image via CPU    */);

Texture*
opengl_upload_mask (GLenum iTarget, /* GL texture-target to use */
		    GLuint uiUnit,  /* GL multi-texture unit to use */
		    GLint iWidth,   /* width of mask texture    */
		    GLint iHeight   /* height of mask texture   */);

Texture*
opengl_upload_text_texture (gchar* pcText, /* text to render into texture    */
			    GLuint uiUnit, /* GL multi-texture unit to use   */
			    GLint iWidth,  /* width of win. text appears in  */
			    GLint iHeight  /* height of win. text appears in */);

void
opengl_delete_texture (Texture* pTexture /* texture to delete */);

void
opengl_paint_checker (GLfloat fStartX, /* x-coord of lower-left corner */
		      GLfloat fStartY, /* y-coord of lower-left corner */
		      GLfloat fSize,   /* width/height of one tile     */
		      GLint iColumns,  /* number of columns to draw    */
		      GLint iRows      /* number of rows to draw       */);

void
opengl_paint_bg (GLfloat fStartX, /* x-coord of lower-left corner */
		 GLfloat fStartY, /* y-coord of lower-left corner */
		 GLfloat fWidth,  /* width of background          */
		 GLfloat fHeight  /* height of background         */);

void
opengl_init (GLint iWidth, /* width of window  */
	     GLint iHeight /* height of window */);

void
opengl_resize (GLsizei iWidth, /* new width to adjust GL-context to  */
	       GLsizei iHeight /* new height to adjust GL-context to */);

void
opengl_draw_textured_quad (Texture* pTexture,   /* texture to use          */
			   GLuint   uiFragProg, /* shader to use           */
			   GLfloat* pfPosition, /* position of quad        */
			   GLfloat  fAngle,     /* current rotation angle  */
			   GLfloat* pfAxis,     /* rotation-axis           */
			   gboolean bNormal,    /* normal or reflected     */
			   GLfloat  fValue      /* value to pass to shader */);

void
opengl_blur_pass (Texture*	pSrcTexture,	/* texture to read from      */
		  Texture*	pDstTexture,	/* texture to write to       */
		  GLuint	uiFragProg,	/* shader to use             */
		  GLfloat	fDeviation,	/* std. deviatino to use     */
		  PassDirection	direction,	/* do which pass of the blur */
		  GLint		iWindowWidth,	/* width of window/context   */
		  GLint		iWindowHeight	/* height of window/context  */);

void
opengl_draw (GLsizei iWidth,  /* current width of the GL-context  */
	     GLsizei iHeight, /* current height of the GL-context */
	     GLfloat fFactor  /* some factor to use for anything  */);

void
opengl_cleanup (void);

#endif /* _OPENGL_H */

