/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** project:
**    OpenGL/cairo anti-aliasing trick
**
** file:
**    shaders.c
**
** author:
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**
** copyright (C) Mirco Mueller, December 2006, placed under the terms of the GPL
**
*******************************************************************************/

#include <gtk/gtk.h>
#include <gtk/gtkgl.h>

#include "shaders.h"

glVertexAttrib1sARBFunc glVertexAttrib1sARB = 0;
glVertexAttrib1fARBFunc glVertexAttrib1fARB = 0;
glVertexAttrib1dARBFunc glVertexAttrib1dARB = 0;
glVertexAttrib2sARBFunc glVertexAttrib2sARB = 0;
glVertexAttrib2fARBFunc glVertexAttrib2fARB = 0;
glVertexAttrib2dARBFunc glVertexAttrib2dARB = 0;
glVertexAttrib3sARBFunc glVertexAttrib3sARB = 0;
glVertexAttrib3fARBFunc glVertexAttrib3fARB = 0;
glVertexAttrib3dARBFunc glVertexAttrib3dARB = 0;
glVertexAttrib4sARBFunc glVertexAttrib4sARB = 0;
glVertexAttrib4fARBFunc glVertexAttrib4fARB = 0;
glVertexAttrib4dARBFunc glVertexAttrib4dARB = 0;
glVertexAttrib4NubARBFunc glVertexAttrib4NubARB = 0;
glVertexAttrib1svARBFunc glVertexAttrib1svARB = 0;
glVertexAttrib1fvARBFunc glVertexAttrib1fvARB = 0;
glVertexAttrib1dvARBFunc glVertexAttrib1dvARB = 0;
glVertexAttrib2svARBFunc glVertexAttrib2svARB = 0;
glVertexAttrib2fvARBFunc glVertexAttrib2fvARB = 0;
glVertexAttrib2dvARBFunc glVertexAttrib2dvARB = 0;
glVertexAttrib3svARBFunc glVertexAttrib3svARB = 0;
glVertexAttrib3fvARBFunc glVertexAttrib3fvARB = 0;
glVertexAttrib3dvARBFunc glVertexAttrib3dvARB = 0;
glVertexAttrib4bvARBFunc glVertexAttrib4bvARB = 0;
glVertexAttrib4svARBFunc glVertexAttrib4svARB = 0;
glVertexAttrib4ivARBFunc glVertexAttrib4ivARB = 0;
glVertexAttrib4ubvARBFunc glVertexAttrib4ubvARB = 0;
glVertexAttrib4usvARBFunc glVertexAttrib4usvARB = 0;
glVertexAttrib4uivARBFunc glVertexAttrib4uivARB = 0;
glVertexAttrib4fvARBFunc glVertexAttrib4fvARB = 0;
glVertexAttrib4dvARBFunc glVertexAttrib4dvARB = 0;
glVertexAttrib4NbvARBFunc glVertexAttrib4NbvARB = 0;
glVertexAttrib4NsvARBFunc glVertexAttrib4NsvARB = 0;
glVertexAttrib4NivARBFunc glVertexAttrib4NivARB = 0;
glVertexAttrib4NubvARBFunc glVertexAttrib4NubvARB = 0;
glVertexAttrib4NusvARBFunc glVertexAttrib4NusvARB = 0;
glVertexAttrib4NuivARBFunc glVertexAttrib4NuivARB = 0;
glVertexAttribPointerARBFunc glVertexAttribPointerARB = 0;
glEnableVertexAttribArrayARBFunc glEnableVertexAttribArrayARB = 0;
glDisableVertexAttribArrayARBFunc glDisableVertexAttribArrayARB = 0;
glProgramStringARBFunc glProgramStringARB = 0;
glBindProgramARBFunc glBindProgramARB = 0;
glDeleteProgramsARBFunc glDeleteProgramsARB = 0;
glGenProgramsARBFunc glGenProgramsARB = 0;
glProgramEnvParameter4dARBFunc glProgramEnvParameter4dARB = 0;
glProgramEnvParameter4dvARBFunc glProgramEnvParameter4dvARB = 0;
glProgramEnvParameter4fARBFunc glProgramEnvParameter4fARB = 0;
glProgramEnvParameter4fvARBFunc glProgramEnvParameter4fvARB = 0;
glProgramLocalParameter4dARBFunc glProgramLocalParameter4dARB = 0;
glProgramLocalParameter4dvARBFunc glProgramLocalParameter4dvARB = 0;
glProgramLocalParameter4fARBFunc glProgramLocalParameter4fARB = 0;
glProgramLocalParameter4fvARBFunc glProgramLocalParameter4fvARB = 0;
glGetProgramEnvParameterdvARBFunc glGetProgramEnvParameterdvARB = 0;
glGetProgramEnvParameterfvARBFunc glGetProgramEnvParameterfvARB = 0;
glGetProgramLocalParameterdvARBFunc glGetProgramLocalParameterdvARB = 0;
glGetProgramLocalParameterfvARBFunc glGetProgramLocalParameterfvARB = 0;
glGetProgramivARBFunc glGetProgramivARB = 0;
glGetProgramStringARBFunc glGetProgramStringARB = 0;
glGetVertexAttribdvARBFunc glGetVertexAttribdvARB = 0;
glGetVertexAttribfvARBFunc glGetVertexAttribfvARB = 0;
glGetVertexAttribivARBFunc glGetVertexAttribivARB = 0;
glGetVertexAttribPointervARBFunc glGetVertexAttribPointervARB = 0;
glIsProgramARBFunc glIsProgramARB = 0;

/* this call checks (and sets up) all functions defined by the OpenGL ARB for...
**
**     GL_ARB_vertex_program
**     GL_ARB_fragment_program
**
** ... which mainly means the capability of an OpenGL implementation to execute
** shader-assembly (_not_ the high-level GLSL, OpenGL shading language) fully
** replacing the old fixed-function pipeline; remember: if you enable this (the
** call returned TRUE) you have to provide equivalents for the fixed-function
** pipeline functionality in shader-assembly, otherwise you won't see much on
** the screen; you cannot mix programmable shading and fixed functions if the
** shading-hardware is enabled */
gboolean
setup_shader_callbacks (void)
{
	gboolean bResult = TRUE;
	const gchar* dummy = NULL;

	dummy = (gchar*) glGetString (GL_EXTENSIONS);
	if (g_strrstr (dummy, "GL_ARB_fragment_program") == NULL ||
	    g_strrstr (dummy, "GL_ARB_vertex_program") == NULL)
	{
		g_print ("Could not find OpenGL ARB shader-program support!\n");
		return FALSE;
	}

	glVertexAttrib1sARB = (glVertexAttrib1sARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib1sARB");
	if (!glVertexAttrib1sARB)
	{
		g_print ("Could not get address of glVertexAttrib1sARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib1fARB = (glVertexAttrib1fARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib1fARB");
	if (!glVertexAttrib1fARB)
	{
		g_print ("Could not get address of glVertexAttrib1fARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib1fARB = (glVertexAttrib1fARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib1fARB");
	if (!glVertexAttrib1fARB)
	{
		g_print ("Could not get address of glVertexAttrib1fARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib1dARB = (glVertexAttrib1dARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib1dARB");
	if (!glVertexAttrib1dARB)
	{
		g_print ("Could not get address of glVertexAttrib1dARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib2sARB = (glVertexAttrib2sARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib2sARB");
	if (!glVertexAttrib2sARB)
	{
		g_print ("Could not get address of glVertexAttrib2sARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib2fARB = (glVertexAttrib2fARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib2fARB");
	if (!glVertexAttrib2fARB)
	{
		g_print ("Could not get address of glVertexAttrib2fARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib2dARB = (glVertexAttrib2dARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib2dARB");
	if (!glVertexAttrib2dARB)
	{
		g_print ("Could not get address of glVertexAttrib2dARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib3sARB = (glVertexAttrib3sARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib3sARB");
	if (!glVertexAttrib3sARB)
	{
		g_print ("Could not get address of glVertexAttrib3sARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib3fARB = (glVertexAttrib3fARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib3fARB");
	if (!glVertexAttrib3fARB)
	{
		g_print ("Could not get address of glVertexAttrib3fARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib3dARB = (glVertexAttrib3dARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib3dARB");
	if (!glVertexAttrib3dARB)
	{
		g_print ("Could not get address of glVertexAttrib3dARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4sARB = (glVertexAttrib4sARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4sARB");
	if (!glVertexAttrib4sARB)
	{
		g_print ("Could not get address of glVertexAttrib4sARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4fARB = (glVertexAttrib4fARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4fARB");
	if (!glVertexAttrib4fARB)
	{
		g_print ("Could not get address of glVertexAttrib4fARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4dARB = (glVertexAttrib4dARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4dARB");
	if (!glVertexAttrib4dARB)
	{
		g_print ("Could not get address of glVertexAttrib4dARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4NubARB = (glVertexAttrib4NubARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4NubARB");
	if (!glVertexAttrib4NubARB)
	{
		g_print ("Could not get address of glVertexAttrib4NubARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib1svARB = (glVertexAttrib1svARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib1svARB");
	if (!glVertexAttrib1svARB)
	{
		g_print ("Could not get address of glVertexAttrib1svARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib1fvARB = (glVertexAttrib1fvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib1fvARB");
	if (!glVertexAttrib1fvARB)
	{
		g_print ("Could not get address of glVertexAttrib1fvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib1dvARB = (glVertexAttrib1dvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib1dvARB");
	if (!glVertexAttrib1dvARB)
	{
		g_print ("Could not get address of glVertexAttrib1dvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib2svARB = (glVertexAttrib2svARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib2svARB");
	if (!glVertexAttrib2svARB)
	{
		g_print ("Could not get address of glVertexAttrib2svARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib2fvARB = (glVertexAttrib2fvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib2fvARB");
	if (!glVertexAttrib2fvARB)
	{
		g_print ("Could not get address of glVertexAttrib2fvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib2dvARB = (glVertexAttrib2dvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib2dvARB");
	if (!glVertexAttrib2dvARB)
	{
		g_print ("Could not get address of glVertexAttrib2dvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib3svARB = (glVertexAttrib3svARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib3svARB");
	if (!glVertexAttrib3svARB)
	{
		g_print ("Could not get address of glVertexAttrib3svARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib3fvARB = (glVertexAttrib3fvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib3fvARB");
	if (!glVertexAttrib3fvARB)
	{
		g_print ("Could not get address of glVertexAttrib3fvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib3dvARB = (glVertexAttrib3dvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib3dvARB");
	if (!glVertexAttrib3dvARB)
	{
		g_print ("Could not get address of glVertexAttrib3dvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4bvARB = (glVertexAttrib4bvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4bvARB");
	if (!glVertexAttrib4bvARB)
	{
		g_print ("Could not get address of glVertexAttrib4bvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4svARB = (glVertexAttrib4svARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4svARB");
	if (!glVertexAttrib4svARB)
	{
		g_print ("Could not get address of glVertexAttrib4svARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4ivARB = (glVertexAttrib4ivARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4ivARB");
	if (!glVertexAttrib4ivARB)
	{
		g_print ("Could not get address of glVertexAttrib4ivARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4ubvARB = (glVertexAttrib4ubvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4ubvARB");
	if (!glVertexAttrib4ubvARB)
	{
		g_print ("Could not get address of glVertexAttrib4ubvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4usvARB = (glVertexAttrib4usvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4usvARB");
	if (!glVertexAttrib4usvARB)
	{
		g_print ("Could not get address of glVertexAttrib4usvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4uivARB = (glVertexAttrib4uivARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4uivARB");
	if (!glVertexAttrib4uivARB)
	{
		g_print ("Could not get address of glVertexAttrib4uivARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4fvARB = (glVertexAttrib4fvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4fvARB");
	if (!glVertexAttrib4fvARB)
	{
		g_print ("Could not get address of glVertexAttrib4fvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4dvARB = (glVertexAttrib4dvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4dvARB");
	if (!glVertexAttrib4dvARB)
	{
		g_print ("Could not get address of glVertexAttrib4dvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4NbvARB = (glVertexAttrib4NbvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4NbvARB");
	if (!glVertexAttrib4NbvARB)
	{
		g_print ("Could not get address of glVertexAttrib4NbvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4NsvARB = (glVertexAttrib4NsvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4NsvARB");
	if (!glVertexAttrib4NsvARB)
	{
		g_print ("Could not get address of glVertexAttrib4NsvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4NivARB = (glVertexAttrib4NivARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4NivARB");
	if (!glVertexAttrib4NivARB)
	{
		g_print ("Could not get address of glVertexAttrib4NivARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4NubvARB = (glVertexAttrib4NubvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4NubvARB");
	if (!glVertexAttrib4NubvARB)
	{
		g_print ("Could not get address of glVertexAttrib4NubvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4NusvARB = (glVertexAttrib4NusvARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4NusvARB");
	if (!glVertexAttrib4NusvARB)
	{
		g_print ("Could not get address of glVertexAttrib4NusvARB()!\n");
		bResult = FALSE;
	}

	glVertexAttrib4NuivARB = (glVertexAttrib4NuivARBFunc)
		gdk_gl_get_proc_address ("glVertexAttrib4NuivARB");
	if (!glVertexAttrib4NuivARB)
	{
		g_print ("Could not get address of glVertexAttrib4NuivARB()!\n");
		bResult = FALSE;
	}

	glVertexAttribPointerARB = (glVertexAttribPointerARBFunc)
		gdk_gl_get_proc_address ("glVertexAttribPointerARB");
	if (!glVertexAttribPointerARB)
	{
		g_print ("Could not get address of glVertexAttribPointerARB()!\n");
		bResult = FALSE;
	}

	glEnableVertexAttribArrayARB = (glEnableVertexAttribArrayARBFunc)
		gdk_gl_get_proc_address ("glEnableVertexAttribArrayARB");
	if (!glEnableVertexAttribArrayARB)
	{
		g_print ("Could not get address of glEnableVertexAttribArrayARB()!\n");
		bResult = FALSE;
	}

	glDisableVertexAttribArrayARB = (glDisableVertexAttribArrayARBFunc)
		gdk_gl_get_proc_address ("glDisableVertexAttribArrayARB");
	if (!glDisableVertexAttribArrayARB)
	{
		g_print ("Could not get address of glDisableVertexAttribArrayARB()!\n");
		bResult = FALSE;
	}

	glProgramStringARB = (glProgramStringARBFunc)
		gdk_gl_get_proc_address ("glProgramStringARB");
	if (!glProgramStringARB)
	{
		g_print ("Could not get address of glProgramStringARB()!\n");
		bResult = FALSE;
	}

	glBindProgramARB = (glBindProgramARBFunc)
		gdk_gl_get_proc_address ("glBindProgramARB");
	if (!glBindProgramARB)
	{
		g_print ("Could not get address of glBindProgramARB()!\n");
		bResult = FALSE;
	}

	glDeleteProgramsARB = (glDeleteProgramsARBFunc)
		gdk_gl_get_proc_address ("glDeleteProgramsARB");
	if (!glDeleteProgramsARB)
	{
		g_print ("Could not get address of glDeleteProgramsARB()!\n");
		bResult = FALSE;
	}

	glGenProgramsARB = (glGenProgramsARBFunc)
		gdk_gl_get_proc_address ("glGenProgramsARB");
	if (!glGenProgramsARB)
	{
		g_print ("Could not get address of glGenProgramsARB()!\n");
		bResult = FALSE;
	}

	glProgramEnvParameter4dARB = (glProgramEnvParameter4dARBFunc)
		gdk_gl_get_proc_address ("glProgramEnvParameter4dARB");
	if (!glProgramEnvParameter4dARB)
	{
		g_print ("Could not get address of glProgramEnvParameter4dARB()!\n");
		bResult = FALSE;
	}

	glProgramEnvParameter4dvARB = (glProgramEnvParameter4dvARBFunc)
		gdk_gl_get_proc_address ("glProgramEnvParameter4dvARB");
	if (!glProgramEnvParameter4dvARB)
	{
		g_print ("Could not get address of glProgramEnvParameter4dvARB()!\n");
		bResult = FALSE;
	}

	glProgramEnvParameter4fARB = (glProgramEnvParameter4fARBFunc)
		gdk_gl_get_proc_address ("glProgramEnvParameter4fARB");
	if (!glProgramEnvParameter4fARB)
	{
		g_print ("Could not get address of glProgramEnvParameter4fARB()!\n");
		bResult = FALSE;
	}

	glProgramEnvParameter4fvARB = (glProgramEnvParameter4fvARBFunc)
		gdk_gl_get_proc_address ("glProgramEnvParameter4fvARB");
	if (!glProgramEnvParameter4fvARB)
	{
		g_print ("Could not get address of glProgramEnvParameter4fvARB()!\n");
		bResult = FALSE;
	}

	glProgramLocalParameter4dARB = (glProgramLocalParameter4dARBFunc)
		gdk_gl_get_proc_address ("glProgramLocalParameter4dARB");
	if (!glProgramLocalParameter4dARB)
	{
		g_print ("Could not get address of glProgramLocalParameter4dARB()!\n");
		bResult = FALSE;
	}

	glProgramLocalParameter4dvARB = (glProgramLocalParameter4dvARBFunc)
		gdk_gl_get_proc_address ("glProgramLocalParameter4dvARB");
	if (!glProgramLocalParameter4dvARB)
	{
		g_print ("Could not get address of glProgramLocalParameter4dvARB()!\n");
		bResult = FALSE;
	}

	glProgramLocalParameter4fARB = (glProgramLocalParameter4fARBFunc)
		gdk_gl_get_proc_address ("glProgramLocalParameter4fARB");
	if (!glProgramLocalParameter4fARB)
	{
		g_print ("Could not get address of glProgramLocalParameter4fARB()!\n");
		bResult = FALSE;
	}

	glProgramLocalParameter4fvARB = (glProgramLocalParameter4fvARBFunc)
		gdk_gl_get_proc_address ("glProgramLocalParameter4fvARB");
	if (!glProgramLocalParameter4fvARB)
	{
		g_print ("Could not get address of glProgramLocalParameter4fvARB()!\n");
		bResult = FALSE;
	}

	glGetProgramEnvParameterdvARB = (glGetProgramEnvParameterdvARBFunc)
		gdk_gl_get_proc_address ("glGetProgramEnvParameterdvARB");
	if (!glGetProgramEnvParameterdvARB)
	{
		g_print ("Could not get address of glGetProgramEnvParameterdvARB()!\n");
		bResult = FALSE;
	}

	glGetProgramEnvParameterfvARB = (glGetProgramEnvParameterfvARBFunc)
		gdk_gl_get_proc_address ("glGetProgramEnvParameterfvARB");
	if (!glGetProgramEnvParameterfvARB)
	{
		g_print ("Could not get address of glGetProgramEnvParameterfvARB()!\n");
		bResult = FALSE;
	}

	glGetProgramLocalParameterdvARB = (glGetProgramLocalParameterdvARBFunc)
		gdk_gl_get_proc_address ("glGetProgramLocalParameterdvARB");
	if (!glGetProgramLocalParameterdvARB)
	{
		g_print ("Could not get address of glGetProgramLocalParameterdvARB()!\n");
		bResult = FALSE;
	}

	glGetProgramLocalParameterfvARB = (glGetProgramLocalParameterfvARBFunc)
		gdk_gl_get_proc_address ("glGetProgramLocalParameterfvARB");
	if (!glGetProgramLocalParameterfvARB)
	{
		g_print ("Could not get address of glGetProgramLocalParameterfvARB()!\n");
		bResult = FALSE;
	}

	glGetProgramivARB = (glGetProgramivARBFunc)
		gdk_gl_get_proc_address ("glGetProgramivARB");
	if (!glGetProgramivARB)
	{
		g_print ("Could not get address of glGetProgramivARB()!\n");
		bResult = FALSE;
	}

	glGetProgramStringARB = (glGetProgramStringARBFunc)
		gdk_gl_get_proc_address ("glGetProgramStringARB");
	if (!glGetProgramStringARB)
	{
		g_print ("Could not get address of glGetProgramStringARB()!\n");
		bResult = FALSE;
	}

	glGetVertexAttribdvARB = (glGetVertexAttribdvARBFunc)
		gdk_gl_get_proc_address ("glGetVertexAttribdvARB");
	if (!glGetVertexAttribdvARB)
	{
		g_print ("Could not get address of glGetVertexAttribdvARB()!\n");
		bResult = FALSE;
	}

	glGetVertexAttribfvARB = (glGetVertexAttribfvARBFunc)
		gdk_gl_get_proc_address ("glGetVertexAttribfvARB");
	if (!glGetVertexAttribfvARB)
	{
		g_print ("Could not get address of glGetVertexAttribfvARB()!\n");
		bResult = FALSE;
	}

	glGetVertexAttribivARB = (glGetVertexAttribivARBFunc)
		gdk_gl_get_proc_address ("glGetVertexAttribivARB");
	if (!glGetVertexAttribivARB)
	{
		g_print ("Could not get address of glGetVertexAttribivARB()!\n");
		bResult = FALSE;
	}

	glGetVertexAttribPointervARB = (glGetVertexAttribPointervARBFunc)
		gdk_gl_get_proc_address ("glGetVertexAttribPointervARB");
	if (!glGetVertexAttribPointervARB)
	{
		g_print ("Could not get address of glGetVertexAttribPointervARB()!\n");
		bResult = FALSE;
	}

	glIsProgramARB = (glIsProgramARBFunc)
		gdk_gl_get_proc_address ("glIsProgramARB");
	if (!glIsProgramARB)
	{
		g_print ("Could not get address of glIsProgramARB()!\n");
		bResult = FALSE;
	}

	return bResult;
}

/* a convenience function for loading a shader from disk and attaching it to a
** shader-program object in GL for esay binding during drawing operations; if
** something fails, an error is printed out to stdout, and a handle of 0 is
** returned; you have to provide the correct program-target iTarget, because
** the shader-program is not examined in advance to automatically select the
** correct program-target for you; FIXME: make the read-buffer-size dynamic */
GLuint
upload_program (GLenum iTarget,   /* either vertex- or fragment-target    */
		gchar* pcFilename /* full path/filename to shader-program */)
{
	GLuint uiProgramId = 0;
	gchar* pcBuffer = NULL;
	const gsize iBufferSize = 16 * 1024; /* FIXME */
	gsize iLength = 0;
	GError* pError = NULL;
	GLint iErrorPos = 0;
	GLint iIsNative = 0;

	/* make sure the filename is valid */
	g_assert (pcFilename != NULL);

	/* allocate new handle for shader-program */
	glGenProgramsARB (1, &uiProgramId);
	if (uiProgramId == 0)
	{
		g_print ("Could not generate shader-handle for \"%s\"!\n",
			 pcFilename);
		return 0;
	}

	/* allocate buffer for shader-source */
	pcBuffer = (gchar*) g_malloc0 (iBufferSize);
	if (!pcBuffer)
	{
		g_print ("Error allocating buffer for \"%s\" shader!\n",
			 pcFilename);
		glDeleteProgramsARB (1, &uiProgramId);
		return 0;
	}

	/* read shader-file into buffer */
	if (!g_file_get_contents (pcFilename,
				  &pcBuffer,
				  &iLength,
				  &pError))
	{
		g_print ("Could read shader-file \"%s\"!\n", pcFilename);
		g_free (pcBuffer);
		glDeleteProgramsARB (1, &uiProgramId);		
		return 0;
	}

	/* upload the shader-source to GL */
	glBindProgramARB (iTarget, uiProgramId);
	glProgramStringARB (iTarget,
			    GL_PROGRAM_FORMAT_ASCII_ARB,
			    iLength,
			    pcBuffer);

	/* free buffer for shader-source */
	g_free (pcBuffer);

	/* get any errors */
	glGetIntegerv (GL_PROGRAM_ERROR_POSITION_ARB, &iErrorPos);
	glGetProgramivARB (iTarget,
			   GL_PROGRAM_UNDER_NATIVE_LIMITS_ARB,
			   &iIsNative);

	/* check for errors */
	if ((iErrorPos == -1) && (iIsNative == 1))
		return uiProgramId;
	else
	{
		g_print ("Syntax/limit-error encountered in shader \"%s\"!\n\n",
			 pcFilename);
		g_print ("Found error:\n\033[31m%s\033[0m\n",
			 glGetString (GL_PROGRAM_ERROR_STRING_ARB));
		glDeleteProgramsARB (1, &uiProgramId);		
		return 0;
	}
}
